#!/usr/bin/env python
'''Run a perceptual diff test locally.

To run:
    source ./common.sh
    ./dpxdt/tools/local_pdiff.py test dpxdt/tools/local_pdiff_demo

This will run the tests described in dpxdt/tools/local_pdiff_demo/*.yaml.
See those files for details.
'''

import copy
import fnmatch
import glob
import json
import logging
import os
import requests
import signal
import subprocess
import sys
import tempfile
import threading
import time
import traceback

# Local Libraries
import gflags
FLAGS = gflags.FLAGS
import pyimgur
import yaml

# Local modules
from dpxdt.client import capture_worker
from dpxdt.client import fetch_worker
from dpxdt.client import pdiff_worker
from dpxdt.client import process_worker
from dpxdt.client import timer_worker
from dpxdt.client import workers

FLAGS.phantomjs_binary = 'phantomjs'
FLAGS.phantomjs_timeout = 20

gflags.DEFINE_boolean(
        'list_tests', False,
        'Set this to list the names of all tests instead of running them.')

gflags.DEFINE_string(
        'test_filter', '',
        'Run a subset of tests. Pass a test name to run just that test, or '
        'use a * to match a set of tests. See '
        'https://code.google.com/p/googletest/wiki/AdvancedGuide'
        '#Running_a_Subset_of_the_Tests for full syntax.')

gflags.DEFINE_string(
        'imgur_client_id', '',
        'When this is set, dpxdt will upload all screenshots from failing '
        'tests to Imgur using their API. This is helpful when running tests '
        'on a Travis-CI worker, for instance. You must register an app with '
        'Imgur to use this.')

MODES = ['test', 'update']

# global tracker
FAILED_TESTS = 0



def should_run_test(name, pattern):
    '''Given a test_filter pattern and a test name, should the test be run?'''
    if pattern == '': return True
    
    def matches_any(name, parts):
        for part in parts:
            if fnmatch.fnmatch(name, part):
                return True
        return False

    positive_negative = pattern.split('-')
    positive = positive_negative[0]
    if positive:
        # There's something here -- have to match it!
        parts = positive.split(':')
        if not matches_any(name, parts): return False

    if len(positive_negative) > 1:
        negative = positive_negative[1]
        parts = negative.split(':')
        if matches_any(name, parts): return False

    return True


class OneTestWorkflowItem(workers.WorkflowItem):
    '''Runs an individual capture & pdiff (or update) based on a config.'''

    def run(self, test_config, ref_dir, tmp_dir, mode, heartbeat=None, num_attempts=0):
        '''Build a CaptureAndDiffWorkflowItem for a test.
        
        Args:
            test_config: See test.yaml for structure of test_config.
        Returns: A CaptureAndDiffWorkflowItem
        '''
        assert 'name' in test_config
        name = test_config['name']

        if 'ref' in test_config:
            # This test has different ref/run arms.
            assert 'run' in test_config
            arm_config = { 'name': name }
            if mode == 'test':
                arm_config.update(test_config['run'])
            elif mode == 'update':
                arm_config.update(test_config['ref'])
            test_config = arm_config

        assert 'url' in test_config

        test_dir = tempfile.mkdtemp(dir=tmp_dir)
        log_file = os.path.join(test_dir, 'log.txt')
        output_path = os.path.join(test_dir, 'screenshot.png')

        logging.info('Test config:\n%s', json.dumps(test_config, indent=2))

        capture_config = copy.deepcopy(test_config['config'])
        capture_config['targetUrl'] = test_config['url']
        config_file = os.path.join(test_dir, 'config.json')
        json.dump(capture_config, open(config_file, 'w'), indent=2)

        ref_path = os.path.join(ref_dir, '%s.png' % name)
        if mode == 'test':
            assert os.path.exists(ref_path), (
                'Reference image %s does not exist. '
                'Try running in update mode.' % ref_path)
        elif mode == 'update':
            output_path = ref_path
            ref_path = None
        else:
            raise ValueError('Invalid mode %s' % mode)

        class NamedHeartbeat(workers.WorkflowItem):
            def run(self, message):
                yield heartbeat('%s: %s' % (name, message))

        try:
            yield CaptureAndDiffWorkflowItem(
                    name, log_file, config_file, output_path, ref_path,
                    heartbeat=NamedHeartbeat)
        except capture_worker.CaptureFailedError, e:
            if num_attempts >= e.max_attempts:
                yield heartbeat('Unable to capture screenshot after %d tries.' % num_attempts)
                raise
            else:
                num_attempts += 1
                yield heartbeat('Capture failed, retrying (%d)' % num_attempts)
                yield OneTestWorkflowItem(test_config, ref_dir, tmp_dir, mode,
                        heartbeat=heartbeat, num_attempts=num_attempts)


class CaptureAndDiffWorkflowItem(workers.WorkflowItem):

    def run(self, name, log_file, config_file, output_path, ref_path, heartbeat=None):
        yield heartbeat('Running webpage capture process')
        yield heartbeat('  Logging to %s' % log_file)

        capture_failed = True
        failure_reason = None

        try:
            returncode = yield capture_worker.CaptureWorkflow(log_file, config_file, output_path)
        except (process_worker.TimeoutError, OSError), e:
            failure_reason = str(e)
        else:
            capture_failed = returncode != 0
            failure_reason = 'returncode=%s' % returncode

        if capture_failed:
            raise capture_worker.CaptureFailedError(
                FLAGS.capture_task_max_attempts,
                failure_reason)

        if ref_path is None:
            yield heartbeat('Updated %s' % output_path)
            return  # update mode
        
        # TODO: consolidate this code w/ DoPdiffQueueWorkflow.run
        ref_resized_path = os.path.join(os.path.dirname(output_path), 'ref_resized')
        diff_path = os.path.join(os.path.dirname(output_path), 'diff.png')
        max_attempts = FLAGS.pdiff_task_max_attempts

        yield heartbeat('Resizing reference image')
        returncode = yield pdiff_worker.ResizeWorkflow(
            log_file, ref_path, output_path, ref_resized_path)
        if returncode != 0:
            raise pdiff_worker.PdiffFailedError(
                max_attempts,
                'Could not resize reference image to size of new image')

        yield heartbeat('Running perceptual diff process')
        returncode = yield pdiff_worker.PdiffWorkflow(
            log_file, ref_resized_path, output_path, diff_path)

        # ImageMagick returns 1 if the images are different and 0 if
        # they are the same, so the return code is a bad judge of
        # successfully running the diff command. Instead we need to check
        # the output text.
        diff_failed = True

        # Check for a successful run or a known failure.
        distortion = None
        if os.path.isfile(log_file):
            log_data = open(log_file).read()
            if 'all: 0 (0)' in log_data:
                diff_path = None
                diff_failed = False
            elif 'image widths or heights differ' in log_data:
                # Give up immediately
                max_attempts = 1
            else:
                # Try to find the image magic normalized root square
                # mean and grab the first one.
                r = pdiff_worker.DIFF_REGEX.findall(log_data)
                if len(r) > 0:
                    diff_failed = False
                    distortion = r[0]

        if diff_failed:
            raise pdiff_worker.PdiffFailedError(
                max_attempts,
                'Comparison failed. returncode=%r' % returncode)

        else:
            if distortion:
                print '%s failed' % name
                print '  %s distortion' % distortion
                print '  Ref:  %s' % self.maybe_imgur(ref_resized_path)
                print '  Run:  %s' % self.maybe_imgur(output_path)
                print '  Diff: %s' % self.maybe_imgur(diff_path)

                # convenience line for copy/pasting
                print ' (all): %s/{%s}' % (
                        os.path.dirname(output_path),
                        ','.join(map(os.path.basename,
                            [ref_resized_path, output_path, diff_path])))
                global FAILED_TESTS
                FAILED_TESTS += 1

            else:
                print '%s passed (no diff)' % name

        # TODO: delete temp files

    def maybe_imgur(self, path):
        '''Uploads a file to imgur if requested via command line flags.
        
        Returns either "path" or "path url" depending on the course of action.
        '''
        if not FLAGS.imgur_client_id:
            return path

        im = pyimgur.Imgur(FLAGS.imgur_client_id)
        uploaded_image = im.upload_image(path)
        return '%s %s' % (path, uploaded_image.link)


class SetupStep(object):
    '''Logic for running and finishing the setup step of a pdiff test.'''

    def __init__(self, config, tmp_dir):
        '''Config is the top-level test config YAML object.'''
        self._config = config
        self._setup = config.get('setup')
        self._tmp_dir = tmp_dir
        self._setup_proc = None

    def run(self):
        if not self._setup: return

        # Note: we cannot use ProcessWorkflow here because the setup script
        # is not expected to terminate (it typically spawns a long-lived server).
        setup_file = os.path.join(self._tmp_dir, 'setup.sh')
        log_file = os.path.join(self._tmp_dir, 'setup.log')
        logging.info('Executing setup step: %s', setup_file)
        open(setup_file, 'w').write(self._setup)

        # If the shell script launches its own subprocesses (e.g. servers),
        # then these will become orphans if we send SIGTERM to setup_proc. In
        # order to avoid this, we make the shell script its own process group.
        # See http://stackoverflow.com/a/4791612/388951
        with open(log_file, 'a') as output_file:
            self._setup_proc = subprocess.Popen(['bash', setup_file],
                stderr=subprocess.STDOUT,
                stdout=output_file,
                close_fds=True,
                preexec_fn=os.setsid)

        return {'script': setup_file, 'log': log_file}

    def terminate(self):
        if not self._setup_proc: return
        if self._setup_proc.pid > 0:
            # TODO: send SIGKILL after 5 seconds?
            os.killpg(self._setup_proc.pid, signal.SIGTERM)
            self._setup_proc.wait()


class WrappedProcessWorkflowItem(process_worker.ProcessWorkflow):
    '''A ProcessWorkflow which can be yielded inline.'''
    def __init__(self, log_path, args, timeout_seconds=30):
        process_worker.ProcessWorkflow.__init__(
            self, log_path, timeout_seconds=timeout_seconds)
        self._args = args

    def get_args(self):
        return self._args


class WaitForUrlWorkflowItem(workers.WorkflowItem):
    '''Waits for an URL to resolve, with a timeout.'''

    def run(self, tmp_dir, waitfor, heartbeat=None, start_time=None):
        assert 'url' in waitfor
        timeout = waitfor.get('timeout_secs', 10)
        if not start_time:
            start_time = time.time()

        class NotReadyError(Exception):
            pass

        try:
            url = waitfor['url']
            r = requests.head(url)
            if r.status_code != 200:
                yield heartbeat('Request for %s failed (%d)' % (url, r.status_code))
                raise NotReadyError()

            yield heartbeat('Request for %s succeeded, continuing with tests...' % url)
            return
        except requests.ConnectionError, NotReadyError:
            now = time.time()
            if now - start_time >= timeout:
                raise process_worker.TimeoutError()
            yield timer_worker.TimerItem(0.5)  # wait 500ms between checks
            yield WaitForUrlWorkflowItem(tmp_dir, waitfor, heartbeat, start_time)


class WaitForWorkflowItem(workers.WorkflowItem):
    '''This performs the "waitFor" step specified in a test config.'''
    def run(self, config, tmp_dir, heartbeat):
        waitfor = config.get('waitFor')
        if isinstance(waitfor, basestring):
            waitfor_file = os.path.join(tmp_dir, 'waitfor.sh')
            log_file = os.path.join(tmp_dir, 'waitfor.log')
            logging.info('Executing waitfor step: %s', waitfor_file)
            try:
                yield WrappedProcessWorkflowItem(log_file, ['bash', waitfor_file])
            except subprocess.CalledProcessError:
                yield heartbeat('waitFor returned error code\nSee %s' % log_file)
                raise

        elif 'url' in waitfor:
            yield WaitForUrlWorkflowItem(tmp_dir, waitfor, heartbeat)


class RunAllTestSuitesWorkflowItem(workers.WorkflowItem):
    '''Load test YAML files and add them to the work queue.'''

    def run(self, config_dir, mode):
        configs = glob.glob(os.path.join(config_dir, '*.yaml'))
        if not configs:
            raise ValueError('No yaml files found in %s' % config_dir)

        heartbeat=workers.PrintWorkflow

        for config_file in configs:
            config = yaml.load(open(config_file))
            assert 'tests' in config

            if FLAGS.list_tests:
                print '%s:' % config_file
                for test in config['tests']:
                    assert 'name' in test
                    print '  %s' % test['name']
            else:
                yield RunTestSuiteWorkflowItem(config_dir, config, mode, heartbeat)


class RunTestSuiteWorkflowItem(workers.WorkflowItem):
    '''Run a single YAML file's worth of tests.'''

    def run(self, config_dir, config, mode, heartbeat):
        tmp_dir = tempfile.mkdtemp()

        yield heartbeat('Running setup step')
        setup = SetupStep(config, tmp_dir)
        setup_files = setup.run()
        yield heartbeat('  logging to %s' % setup_files['log'])

        try:
            if config.get('waitFor'):
                try:
                    yield WaitForWorkflowItem(config, tmp_dir, heartbeat)
                except process_worker.TimeoutError:
                    # The raw exception has an excessively long stack trace.
                    # This at least adds some helpful context to the end.
                    sys.stderr.write('Timed out on waitFor step.\n')
                    return

            for test in config['tests']:
                assert 'name' in test
                name = test['name']
                if should_run_test(name, FLAGS.test_filter):
                    yield OneTestWorkflowItem(test, config_dir, tmp_dir, mode,
                        heartbeat=heartbeat)
                else:
                    logging.info('Skipping %s due to --test_filter=%s',
                            name, FLAGS.test_filter)
        finally:
            setup.terminate()  # kill server from the setup step.


class RepetitiveLogFilterer(object):
    '''Suppress repeated log entries from the same line in the same file.'''
    def __init__(self):
        self.last_source = None

    def filter(self, record):
        if FLAGS.verbose:
            return True
        source = '%s:%s' % (record.filename, record.lineno)
        if source == self.last_source:
            return False
        self.last_source = source

        return True


class CompactExceptionLogger(logging.Formatter):
    def formatException(self, ei):
        # Like logging.Formatter.formatException, but without the stack trace.
        if FLAGS.verbose:
            return super(CompactExceptionLogger, self).formatException(ei)
        else:
            return '\n'.join(traceback.format_exception_only(ei[0], ei[1]))


def usage(short=False):
    sys.stderr.write('Usage: %s [update|test] <testdir>\n' % sys.argv[0])
    if not short:
        sys.stderr.write('%s\n' % FLAGS)


def main(argv):
    try:
        argv = FLAGS(argv)
    except gflags.FlagsError, e:
        sys.stderr.write('%s\n' % e)
        usage()
        sys.exit(1)

    if len(argv) < 3:
        sys.stderr.write('Too few arguments\n')
        usage(short=True)
        sys.exit(1)

    mode = argv[1]
    assert mode in MODES, 'Invalid mode: %s (expected %r)' % (mode, MODES)

    config_dir = argv[2]
    assert os.path.isdir(config_dir), 'Expected directory, got %s' % config_dir

    assert os.path.exists(FLAGS.phantomjs_script)

    logging.basicConfig()
    logging.getLogger().addFilter(RepetitiveLogFilterer())
    logging.getLogger().handlers[0].setFormatter(CompactExceptionLogger())
    if FLAGS.verbose:
        logging.getLogger().setLevel(logging.DEBUG)

    coordinator = workers.get_coordinator()
    timer_worker.register(coordinator)

    global FAILED_TESTS
    FAILED_TESTS = 0
    item = RunAllTestSuitesWorkflowItem(config_dir, mode)
    item.root = True
    coordinator.input_queue.put(item, mode)

    coordinator.start()
    coordinator.wait_one()
    coordinator.stop()
    coordinator.join()

    if mode == 'test':
        if FAILED_TESTS > 0:
            sys.stderr.write('%d test(s) failed.\n' % FAILED_TESTS)
            sys.exit(1)
        else:
            sys.stderr.write('All tests passed!\n')
            sys.exit(0)


def run():
    # (intended to be run from package)

    FLAGS.phantomjs_script = os.path.join(
            os.path.dirname(__file__), '..', 'client', 'capture.js')

    main(sys.argv)


if __name__ == '__main__':
    run()



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pull-queue web handlers."""

import logging

# Local libraries
import flask
from flask import Flask, redirect, render_template, request, url_for
from sqlalchemy import func

# Local modules
from . import app
from . import db
from dpxdt.server import auth
from dpxdt.server import forms
from dpxdt.server import utils
from dpxdt.server import work_queue


@app.route('/api/work_queue/<string:queue_name>/add', methods=['POST'])
@auth.superuser_api_key_required
@utils.retryable_transaction()
def handle_add(queue_name):
    """Adds a task to a queue."""
    source = request.form.get('source', request.remote_addr, type=str)
    try:
        task_id = work_queue.add(
            queue_name,
            payload=request.form.get('payload', type=str),
            content_type=request.form.get('content_type', type=str),
            source=source,
            task_id=request.form.get('task_id', type=str))
    except work_queue.Error, e:
        return utils.jsonify_error(e)

    db.session.commit()
    logging.info('Task added: queue=%r, task_id=%r, source=%r',
                 queue_name, task_id, source)
    return flask.jsonify(task_id=task_id)


@app.route('/api/work_queue/<string:queue_name>/lease', methods=['POST'])
@auth.superuser_api_key_required
@utils.retryable_transaction()
def handle_lease(queue_name):
    """Leases a task from a queue."""
    owner = request.form.get('owner', request.remote_addr, type=str)
    try:
        task_list = work_queue.lease(
            queue_name,
            owner,
            request.form.get('count', 1, type=int),
            request.form.get('timeout', 60, type=int))
    except work_queue.Error, e:
        return utils.jsonify_error(e)

    if not task_list:
        return flask.jsonify(tasks=[])

    db.session.commit()
    task_ids = [t['task_id'] for t in task_list]
    logging.debug('Task leased: queue=%r, task_ids=%r, owner=%r',
                  queue_name, task_ids, owner)
    return flask.jsonify(tasks=task_list)


@app.route('/api/work_queue/<string:queue_name>/heartbeat', methods=['POST'])
@auth.superuser_api_key_required
@utils.retryable_transaction()
def handle_heartbeat(queue_name):
    """Updates the heartbeat message for a task."""
    task_id = request.form.get('task_id', type=str)
    message = request.form.get('message', type=str)
    index = request.form.get('index', type=int)
    try:
        work_queue.heartbeat(
            queue_name,
            task_id,
            request.form.get('owner', request.remote_addr, type=str),
            message,
            index)
    except work_queue.Error, e:
        return utils.jsonify_error(e)

    db.session.commit()
    logging.debug('Task heartbeat: queue=%r, task_id=%r, message=%r, index=%d',
                  queue_name, task_id, message, index)
    return flask.jsonify(success=True)


@app.route('/api/work_queue/<string:queue_name>/finish', methods=['POST'])
@auth.superuser_api_key_required
@utils.retryable_transaction()
def handle_finish(queue_name):
    """Marks a task on a queue as finished."""
    task_id = request.form.get('task_id', type=str)
    owner = request.form.get('owner', request.remote_addr, type=str)
    error = request.form.get('error', type=str) is not None
    try:
        work_queue.finish(queue_name, task_id, owner, error=error)
    except work_queue.Error, e:
        return utils.jsonify_error(e)

    db.session.commit()
    logging.debug('Task finished: queue=%r, task_id=%r, owner=%r, error=%r',
                  queue_name, task_id, owner, error)
    return flask.jsonify(success=True)


@app.route('/api/work_queue')
@auth.superuser_required
def view_all_work_queues():
    """Page for viewing the index of all active work queues."""
    count_list = list(
        db.session.query(
            work_queue.WorkQueue.queue_name,
            work_queue.WorkQueue.status,
            func.count(work_queue.WorkQueue.task_id))
        .group_by(work_queue.WorkQueue.queue_name,
                  work_queue.WorkQueue.status))

    queue_dict = {}
    for name, status, count in count_list:
        queue_dict[(name, status)] = dict(
            name=name, status=status, count=count)

    max_created_list = list(
        db.session.query(
            work_queue.WorkQueue.queue_name,
            work_queue.WorkQueue.status,
            func.max(work_queue.WorkQueue.created))
        .group_by(work_queue.WorkQueue.queue_name,
                  work_queue.WorkQueue.status))

    for name, status, newest_created in max_created_list:
        queue_dict[(name, status)]['newest_created'] = newest_created

    min_eta_list = list(
        db.session.query(
            work_queue.WorkQueue.queue_name,
            work_queue.WorkQueue.status,
            func.min(work_queue.WorkQueue.eta))
        .group_by(work_queue.WorkQueue.queue_name,
                  work_queue.WorkQueue.status))

    for name, status, oldest_eta in min_eta_list:
        queue_dict[(name, status)]['oldest_eta'] = oldest_eta

    queue_list = list(queue_dict.values())
    queue_list.sort(key=lambda x: (x['name'], x['status']))

    context = dict(
        queue_list=queue_list,
    )
    return render_template('view_work_queue_index.html', **context)


@app.route('/api/work_queue/<string:queue_name>', methods=['GET', 'POST'])
@auth.superuser_required
def manage_work_queue(queue_name):
    """Page for viewing the contents of a work queue."""
    modify_form = forms.ModifyWorkQueueTaskForm()
    if modify_form.validate_on_submit():
        primary_key = (modify_form.task_id.data, queue_name)
        task = work_queue.WorkQueue.query.get(primary_key)
        if task:
            logging.info('Action: %s task_id=%r',
                         modify_form.action.data, modify_form.task_id.data)
            if modify_form.action.data == 'retry':
                task.status = work_queue.WorkQueue.LIVE
                task.lease_attempts = 0
                task.heartbeat = 'Retrying ...'
                db.session.add(task)
            else:
                db.session.delete(task)
            db.session.commit()
        else:
            logging.warning('Could not find task_id=%r to delete',
                            modify_form.task_id.data)
        return redirect(url_for('manage_work_queue', queue_name=queue_name))

    query = (
        work_queue.WorkQueue.query
        .filter_by(queue_name=queue_name)
        .order_by(work_queue.WorkQueue.created.desc()))

    status = request.args.get('status', '', type=str).lower()
    if status in work_queue.WorkQueue.STATES:
        query = query.filter_by(status=status)
    else:
        status = None

    item_list = list(query.limit(100))
    work_list = []
    for item in item_list:
        form = forms.ModifyWorkQueueTaskForm()
        form.task_id.data = item.task_id
        form.delete.data = True
        work_list.append((item, form))

    context = dict(
        queue_name=queue_name,
        status=status,
        work_list=work_list,
    )
    return render_template('view_work_queue.html', **context)



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Runs a dpxdt queue worker."""

import Queue
import logging
import sys
import threading

# Local Libraries
import gflags
FLAGS = gflags.FLAGS

# Local modules
from dpxdt.client import capture_worker
from dpxdt.client import fetch_worker
from dpxdt.client import pdiff_worker
from dpxdt.client import timer_worker
from dpxdt.client import workers


def run_workers():
    coordinator = workers.get_coordinator()
    capture_worker.register(coordinator)
    fetch_worker.register(coordinator)
    pdiff_worker.register(coordinator)
    timer_worker.register(coordinator)
    coordinator.start()
    return coordinator
    logging.info('Workers started')


def main(argv):
    try:
        argv = FLAGS(argv)
    except gflags.FlagsError, e:
        print '%s\nUsage: %s ARGS\n%s' % (e, sys.argv[0], FLAGS)
        sys.exit(1)

    logging.basicConfig(
        format='%(levelname)s %(filename)s:%(lineno)s] %(message)s')

    if FLAGS.verbose:
        logging.getLogger().setLevel(logging.DEBUG)

    if FLAGS.verbose_queries:
        logging.getLogger('sqlalchemy.engine').setLevel(logging.DEBUG)

    coordinator = run_workers()
    coordinator.wait_one()
    coordinator.stop()
    coordinator.join()


if __name__ == '__main__':
    main(sys.argv)



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Background worker that does perceptual diffs, possibly from a queue."""

import Queue
import json
import os
import shutil
import subprocess
import sys
import tempfile
import threading
import time
import urllib2
import re

# Local Libraries
import gflags
FLAGS = gflags.FLAGS

# Local modules
from dpxdt import constants
from dpxdt.client import process_worker
from dpxdt.client import queue_worker
from dpxdt.client import release_worker
from dpxdt.client import workers


gflags.DEFINE_integer(
    'pdiff_task_max_attempts', 3,
    'Maximum number of attempts for processing a pdiff task.')

gflags.DEFINE_integer(
    'pdiff_wait_seconds', 3,
    'Wait this many seconds between repeated invocations of pdiff '
    'subprocesses. Can be used to spread out load on the server.')

gflags.DEFINE_string(
    'pdiff_compare_binary', 'compare',
    'Path to the compare binary used for generating perceptual diffs.')

gflags.DEFINE_string(
    'pdiff_composite_binary', 'composite',
    'Path to the composite binary used for resizing images.')

gflags.DEFINE_integer(
    'pdiff_threads', 1, 'Number of perceptual diff threads to run')

gflags.DEFINE_integer(
    'pdiff_timeout', 60,
    'Seconds until we should give up on a pdiff sub-process and try again.')

DIFF_REGEX = re.compile(".*all:.*\(([0-9e\-\.]*)\).*")


class PdiffFailedError(queue_worker.GiveUpAfterAttemptsError):
    """Running a perceptual diff failed for some reason."""


class ResizeWorkflow(process_worker.ProcessWorkflow):
    """Workflow for making images to be diffed the same size."""

    def __init__(self, log_path, ref_path, run_path, resized_ref_path):
        """Initializer.

        Args:
            log_path: Where to write the verbose logging output.
            ref_path: Path to reference screenshot to diff.
            run_path: Path to the most recent run screenshot to diff.
            resized_ref_path: Where the resized ref image should be written.
        """
        process_worker.ProcessWorkflow.__init__(
            self, log_path, timeout_seconds=FLAGS.pdiff_timeout)
        self.ref_path = ref_path
        self.run_path = run_path
        self.resized_ref_path = resized_ref_path

    def get_args(self):
        return [
            FLAGS.pdiff_composite_binary,
            '-compose',
            'src',
            '-gravity',
            'NorthWest',
            self.ref_path,
            self.run_path,
            self.resized_ref_path,
        ]


class PdiffWorkflow(process_worker.ProcessWorkflow):
    """Workflow for doing perceptual diffs using pdiff."""

    def __init__(self, log_path, ref_path, run_path, output_path):
        """Initializer.

        Args:
            log_path: Where to write the verbose logging output.
            ref_path: Path to reference screenshot to diff.
            run_path: Path to the most recent run screenshot to diff.
            output_path: Where the diff image should be written, if any.
        """
        process_worker.ProcessWorkflow.__init__(
            self, log_path, timeout_seconds=FLAGS.pdiff_timeout)
        self.ref_path = ref_path
        self.run_path = run_path
        self.output_path = output_path

    def get_args(self):
        # Method from http://www.imagemagick.org/Usage/compare/
        return [
            FLAGS.pdiff_compare_binary,
            '-verbose',
            '-metric',
            'RMSE',
            '-highlight-color',
            'Red',
            '-compose',
            'Src',
            self.ref_path,
            self.run_path,
            self.output_path,
        ]


class DoPdiffQueueWorkflow(workers.WorkflowItem):
    """Runs the perceptual diff from queue parameters.

    Args:
        build_id: ID of the build.
        release_name: Name of the release.
        release_number: Number of the release candidate.
        run_name: Run to run perceptual diff for.
        reference_sha1sum: Content hash of the previously good image.
        run_sha1sum: Content hash of the new image.
        heartbeat: Function to call with progress status.

    Raises:
        PdiffFailedError if the perceptual diff process failed.
    """

    def run(self, build_id=None, release_name=None, release_number=None,
            run_name=None, reference_sha1sum=None, run_sha1sum=None,
            heartbeat=None):
        output_path = tempfile.mkdtemp()
        try:
            ref_path = os.path.join(output_path, 'ref')
            ref_resized_path = os.path.join(output_path, 'ref_resized')
            run_path = os.path.join(output_path, 'run')
            diff_path = os.path.join(output_path, 'diff.png')
            log_path = os.path.join(output_path, 'log.txt')

            yield heartbeat('Fetching reference and run images')
            yield [
                release_worker.DownloadArtifactWorkflow(
                    build_id, reference_sha1sum, result_path=ref_path),
                release_worker.DownloadArtifactWorkflow(
                    build_id, run_sha1sum, result_path=run_path)
            ]

            max_attempts = FLAGS.pdiff_task_max_attempts

            yield heartbeat('Resizing reference image')
            returncode = yield ResizeWorkflow(
                log_path, ref_path, run_path, ref_resized_path)
            if returncode != 0:
                raise PdiffFailedError(
                    max_attempts,
                    'Could not resize reference image to size of new image')

            yield heartbeat('Running perceptual diff process')
            returncode = yield PdiffWorkflow(
                log_path, ref_resized_path, run_path, diff_path)

            # ImageMagick returns 1 if the images are different and 0 if
            # they are the same, so the return code is a bad judge of
            # successfully running the diff command. Instead we need to check
            # the output text.
            diff_failed = True

            # Check for a successful run or a known failure.
            distortion = None
            if os.path.isfile(log_path):
                log_data = open(log_path).read()
                if 'all: 0 (0)' in log_data:
                    diff_path = None
                    diff_failed = False
                elif 'image widths or heights differ' in log_data:
                    # Give up immediately
                    max_attempts = 1
                else:
                    # Try to find the image magic normalized root square
                    # mean and grab the first one.
                    r = DIFF_REGEX.findall(log_data)
                    if len(r) > 0:
                        diff_failed = False
                        distortion = r[0]

            yield heartbeat('Reporting diff result to server')
            yield release_worker.ReportPdiffWorkflow(
                build_id, release_name, release_number, run_name,
                diff_path, log_path, diff_failed, distortion)

            if diff_failed:
                raise PdiffFailedError(
                    max_attempts,
                    'Comparison failed. returncode=%r' % returncode)
        finally:
            shutil.rmtree(output_path, True)


def register(coordinator):
    """Registers this module as a worker with the given coordinator."""
    assert FLAGS.pdiff_threads > 0
    assert FLAGS.queue_server_prefix

    item = queue_worker.RemoteQueueWorkflow(
        constants.PDIFF_QUEUE_NAME,
        DoPdiffQueueWorkflow,
        max_tasks=FLAGS.pdiff_threads,
        wait_seconds=FLAGS.pdiff_wait_seconds)
    item.root = True
    coordinator.input_queue.put(item)



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common flags for utility scripts."""

# Local Libraries
import gflags
FLAGS = gflags.FLAGS

gflags.DEFINE_string(
    'upload_build_id', None,
    'ID of the build to upload this screenshot set to as a new release.')

gflags.DEFINE_string(
    'upload_release_name', None,
    'Along with upload_build_id, the name of the release to upload to. If '
    'not supplied, a new release will be created.')

gflags.DEFINE_string(
    'inject_css', None,
    'CSS to inject into all captured pages after they have loaded but '
    'before screenshotting.')

gflags.DEFINE_string(
    'inject_js', None,
    'JavaScript to inject into all captured pages after they have loaded '
    'but before screenshotting.')

gflags.DEFINE_string(
    'cookies', None,
    'Filename containing a JSON array of cookies to set.')

gflags.DEFINE_string(
    'release_cut_url', None,
    'URL that describes the release that you are testing. Usually a link to '
    'the commit or branch that was built.')

gflags.DEFINE_string(
    'tests_json_path', None,
    'Path to the JSON file containing the list of tests to diff.')



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Background worker that screenshots URLs, possibly from a queue."""

import Queue
import json
import os
import shutil
import subprocess
import sys
import threading
import tempfile
import time
import urllib2

# Local Libraries
import gflags
FLAGS = gflags.FLAGS

# Local modules
from dpxdt import constants
from dpxdt.client import process_worker
from dpxdt.client import queue_worker
from dpxdt.client import release_worker
from dpxdt.client import workers


gflags.DEFINE_integer(
    'capture_threads', 5, 'Number of website screenshot threads to run')

gflags.DEFINE_integer(
    'capture_task_max_attempts', 3,
    'Maximum number of attempts for processing a capture task.')

gflags.DEFINE_integer(
    'capture_wait_seconds', 3,
    'Wait this many seconds between repeated invocations of capture '
    'subprocesses. Can be used to spread out load on the server.')

gflags.DEFINE_string(
    'phantomjs_binary', None, 'Path to the phantomjs binary')

gflags.DEFINE_string(
    'phantomjs_script', None,
    'Path to the script that drives the phantomjs process')

gflags.DEFINE_integer(
    'phantomjs_timeout', 120,
    'Seconds until giving up on a phantomjs sub-process and trying again.')



class CaptureFailedError(queue_worker.GiveUpAfterAttemptsError):
    """Capturing a webpage screenshot failed for some reason."""


class CaptureWorkflow(process_worker.ProcessWorkflow):
    """Workflow for capturing a website screenshot using PhantomJs."""

    def __init__(self, log_path, config_path, output_path):
        """Initializer.

        Args:
            log_path: Where to write the verbose logging output.
            config_path: Path to the screenshot config file to pass
                to PhantomJs.
            output_path: Where the output screenshot should be written.
        """
        process_worker.ProcessWorkflow.__init__(
            self, log_path, timeout_seconds=FLAGS.phantomjs_timeout)
        self.config_path = config_path
        self.output_path = output_path

    def get_args(self):
        return [
            FLAGS.phantomjs_binary,
            '--disk-cache=false',
            '--debug=true',
            '--ignore-ssl-errors=true',
            FLAGS.phantomjs_script,
            self.config_path,
            self.output_path,
        ]


class DoCaptureQueueWorkflow(workers.WorkflowItem):
    """Runs a webpage screenshot process from queue parameters.

    Args:
        build_id: ID of the build.
        release_name: Name of the release.
        release_number: Number of the release candidate.
        run_name: Run to run perceptual diff for.
        url: URL of the content to screenshot.
        config_sha1sum: Content hash of the config for the new screenshot.
        baseline: Optional. When specified and True, this capture is for
            the reference baseline of the specified run, not the new capture.
        heartbeat: Function to call with progress status.

    Raises:
        CaptureFailedError if the screenshot process failed.
    """

    def run(self, build_id=None, release_name=None, release_number=None,
            run_name=None, url=None, config_sha1sum=None, baseline=None,
            heartbeat=None):
        output_path = tempfile.mkdtemp()
        try:
            image_path = os.path.join(output_path, 'capture.png')
            log_path = os.path.join(output_path, 'log.txt')
            config_path = os.path.join(output_path, 'config.json')
            capture_failed = True
            failure_reason = None

            yield heartbeat('Fetching webpage capture config')
            yield release_worker.DownloadArtifactWorkflow(
                build_id, config_sha1sum, result_path=config_path)

            yield heartbeat('Running webpage capture process')
            try:
                returncode = yield CaptureWorkflow(
                    log_path, config_path, image_path)
            except (process_worker.TimeoutError, OSError), e:
                failure_reason = str(e)
            else:
                capture_failed = returncode != 0
                failure_reason = 'returncode=%s' % returncode

            # Don't upload bad captures, but always upload the error log.
            if capture_failed:
                image_path = None

            yield heartbeat('Reporting capture status to server')
            yield release_worker.ReportRunWorkflow(
                build_id, release_name, release_number, run_name,
                image_path=image_path, log_path=log_path, baseline=baseline,
                run_failed=capture_failed)

            if capture_failed:
                raise CaptureFailedError(
                    FLAGS.capture_task_max_attempts,
                    failure_reason)
        finally:
            shutil.rmtree(output_path, True)


def register(coordinator):
    """Registers this module as a worker with the given coordinator."""
    assert FLAGS.phantomjs_binary
    assert FLAGS.phantomjs_script
    assert FLAGS.capture_threads > 0
    assert FLAGS.queue_server_prefix

    item = queue_worker.RemoteQueueWorkflow(
        constants.CAPTURE_QUEUE_NAME,
        DoCaptureQueueWorkflow,
        max_tasks=FLAGS.capture_threads,
        wait_seconds=FLAGS.capture_wait_seconds)
    item.root = True
    coordinator.input_queue.put(item)



from setuptools import setup, find_packages

setup(name='dpxdt',
      version='0.1.5',
      description='Screenshot diff tool',
      author='Brett Slatkin',
      author_email='brett@haxor.com',
      url='https://github.com/bslatkin/dpxdt/',
      entry_points={
          'console_scripts': [
              'dpxdt = dpxdt.tools.local_pdiff:run',
          ],
      },
      packages=find_packages(exclude=['tests*','dependencies*']),
      install_requires=[
          'PyYAML',
          'python-gflags',
          'poster',
          'pyimgur',
          'requests'
      ],
      include_package_data=True,
      classifiers=[
          'Environment :: Console',
          'Environment :: Web Environment',
          'Framework :: Flask',
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: Apache Software License',
          'Topic :: Software Development :: Version Control'
      ],
)



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Workers for driving screen captures, perceptual diffs, and related work."""

import Queue
import logging
import sys
import threading

# Local Libraries
import gflags
FLAGS = gflags.FLAGS


gflags.DEFINE_float(
    'polltime', 1.0,
    'How long to sleep between polling for work from an input queue, '
    'a subprocess, or a waiting timer.')


class WorkItem(object):
    """Base work item that can be handled by a worker thread."""

    # Instance variables. May be overridden by a sub-class @property.
    error = None
    done = False
    parent = None

    # Set this to True for WorkItems that should never wait for their
    # return values.
    fire_and_forget = False

    def __init__(self):
        pass

    @staticmethod
    def _print_tree(obj):
        if isinstance(obj, dict):
            result = []
            for key, value in obj.iteritems():
                result.append('%s: %s' % (key, WorkItem._print_tree(value)))
            return '{%s}' % ', '.join(result)
        else:
            value_str = repr(obj)
            if len(value_str) > 100:
                return '%s...%s' % (value_str[:100], value_str[-1])
            else:
                return value_str

    def _get_dict_for_repr(self):
        return self.__dict__

    def __repr__(self):
        return '%s.%s(%s)#%d' % (
            self.__class__.__module__,
            self.__class__.__name__,
            self._print_tree(self._get_dict_for_repr()),
            id(self))

    def check_result(self):
        # TODO: For WorkflowItems, remove generator.throw(*item.error) from
        # the stack trace since it's noise. General approach outlined here:
        # https://github.com/mitsuhiko/jinja2/blob/master/jinja2/debug.py
        if self.error:
            raise self.error[0], self.error[1], self.error[2]


class WorkerThread(threading.Thread):
    """Base worker thread that handles items one at a time."""

    def __init__(self, input_queue, output_queue):
        """Initializer.

        Args:
            input_queue: Queue this worker consumes work from.
            output_queue: Queue where this worker puts new work items, if any.
        """
        threading.Thread.__init__(self)
        self.daemon = True
        self.input_queue = input_queue
        self.output_queue = output_queue
        self.interrupted = False
        self.polltime = FLAGS.polltime

    def stop(self):
        """Stops the thread but does not join it."""
        if self.interrupted:
            return
        self.interrupted = True

    def run(self):
        while not self.interrupted:
            try:
                item = self.input_queue.get(True, self.polltime)
            except Queue.Empty:
                self.handle_nothing()
                continue

            try:
                next_item = self.handle_item(item)
            except Exception, e:
                item.error = sys.exc_info()
                logging.exception('%s error item=%r', self.worker_name, item)
                self.output_queue.put(item)
            else:
                logging.debug('%s processed item=%r', self.worker_name, item)
                if not isinstance(item, WorkflowItem):
                    item.done = True
                if next_item:
                    self.output_queue.put(next_item)
            finally:
                self.input_queue.task_done()

    @property
    def worker_name(self):
        return '%s:%s' % (self.__class__.__name__, self.ident)

    def handle_nothing(self):
        """Runs whenever there are no items in the queue."""
        pass

    def handle_item(self, item):
        """Handles a single item.

        Args:
            item: WorkItem to process.

        Returns:
            A WorkItem that should go on the output queue. If None, then
            the provided work item is considered finished and no
            additional work is needed.
        """
        raise NotImplemented


class WorkflowItem(WorkItem):
    """Work item for coordinating other work items.

    To use: Sub-class and override run(). Yield WorkItems you want processed
    as part of this workflow. Exceptions in child workflows will be reinjected
    into the run() generator at the yield point. Results will be available on
    the WorkItems returned by yield statements. Yield a list of WorkItems
    to do them in parallel. The first error encountered for the whole list
    will be raised if there's an exception.
    """

    # Allow these value to be assigned or overridden by a sub-class @property.
    result = None
    root = False

    def __init__(self, *args, **kwargs):
        WorkItem.__init__(self)
        self.args = args
        self.kwargs = kwargs

    def run(self, *args, **kwargs):
        raise NotImplemented


class WaitAny(object):
    """Return control to a workflow after any one of these items finishes.

    As soon as a single work item completes, the combined barrier will be
    fulfilled and control will return to the WorkflowItem. The return values
    will be WorkItem instances, the same ones passed into WaitAny. For
    WorkflowItems the return values will be WorkflowItems if the work is not
    finished yet, and the final return value once work is finished.
    """

    def __init__(self, items):
        """Initializer.

        Args:
            items: List of WorkItems to wait for.
        """
        self.items = items


class Barrier(list):
    """Barrier for running multiple WorkItems in parallel."""

    def __init__(self, workflow, generator, work):
        """Initializer.

        Args:
            workflow: WorkflowItem instance this is for.
            generator: Current state of the WorkflowItem's generator.
            work: Next set of work to do. May be a single WorkItem object or
                a list or tuple that contains a set of WorkItems to run in
                parallel.
        """
        list.__init__(self)
        self.workflow = workflow
        self.generator = generator

        if isinstance(work, (list, tuple)):
            self[:] = list(work)
            self.was_list = True
            self.wait_any = False
        elif isinstance(work, WaitAny):
            self[:] = list(work.items)
            self.was_list = True
            self.wait_any = True
        else:
            self[:] = [work]
            self.was_list = False
            self.wait_any = False

        for item in self:
            assert isinstance(item, WorkItem)
            item.parent = workflow

    @property
    def outstanding(self):
        """Returns whether or not this barrier has pending work."""
        # Allow the same WorkItem to be yielded multiple times but not
        # count towards blocking the barrier.
        done_count = 0
        for item in self:
            if not self.wait_any and item.fire_and_forget:
                # Only count fire_and_forget items as done if this is
                # *not* a WaitAny barrier. We only want to return control
                # to the caller when at least one of the blocking items
                # has completed.
                done_count += 1
            elif item.done:
                done_count += 1

        if self.wait_any and done_count > 0:
            return False

        if done_count == len(self):
            return False

        return True

    @property
    def error(self):
        """Returns the error for this barrier and all work items, if any."""
        # Copy the error from any failed item to be the error for the whole
        # barrier. The first error seen "wins". Also handles the case where
        # the WorkItems passed into the barrier have already completed and
        # been marked with errors.
        for item in self:
            if isinstance(item, WorkItem) and item.error:
                return item.error
        return None

    def get_item(self):
        """Returns the item to send back into the workflow generator."""
        if self.was_list:
            blocking_items = self[:]
            self[:] = []
            for item in blocking_items:
                if (isinstance(item, WorkflowItem) and
                        item.done and
                        not item.error):
                    self.append(item.result)
                else:
                    self.append(item)
            return self
        else:
            return self[0]


class Return(Exception):
    """Raised in WorkflowItem.run to return a result to the caller."""

    def __init__(self, result=None):
        """Initializer.

        Args:
            result: Result of a WorkflowItem, if any.
        """
        self.result = result


class WorkflowThread(WorkerThread):
    """Worker thread for running workflows."""

    def __init__(self, input_queue, output_queue):
        """Initializer.

        Args:
            input_queue: Queue this worker consumes work from. These should be
                WorkflowItems to process, or any WorkItems registered with this
                class using the register() method.
            output_queue: Queue where this worker puts finished work items,
                if any.
        """
        WorkerThread.__init__(self, input_queue, output_queue)
        self.pending = {}
        self.work_map = {}
        self.worker_threads = []
        self.register(WorkflowItem, input_queue)

    # TODO: Implement drain, to let all existing work finish but no new work
    # allowed at the top of the funnel.

    def start(self):
        """Starts the coordinator thread and all related worker threads."""
        assert not self.interrupted
        for thread in self.worker_threads:
            thread.start()
        WorkerThread.start(self)

    def stop(self):
        """Stops the coordinator thread and all related threads."""
        if self.interrupted:
            return
        for thread in self.worker_threads:
            thread.interrupted = True
        self.interrupted = True

    def join(self):
        """Joins the coordinator thread and all worker threads."""
        for thread in self.worker_threads:
            thread.join()
        WorkerThread.join(self)

    def wait_one(self):
        """Waits until this worker has finished one work item or died."""
        while True:
            try:
                item = self.output_queue.get(True, 1)
            except Queue.Empty:
                continue
            except KeyboardInterrupt:
                logging.debug('Exiting')
                return
            else:
                item.check_result()
                return

    def register(self, work_type, queue):
        """Registers where work for a specific type can be executed.

        Args:
            work_type: Sub-class of WorkItem to register.
            queue: Queue instance where WorkItems of the work_type should be
                enqueued when they are yielded by WorkflowItems being run by
                this worker.
        """
        self.work_map[work_type] = queue

    def enqueue_barrier(self, barrier):
        for item in barrier:
            if item.done:
                # Don't reenqueue items that are already done.
                continue

            # Try to find a queue by the work item's class, then by its
            # super class, and so on.
            next_types = [type(item)]
            while next_types:
                try_types = next_types[:]
                next_types[:] = []
                for current_type in try_types:
                    target_queue = self.work_map.get(current_type)
                    if target_queue:
                        next_types = []
                        break
                    next_types.extend(current_type.__bases__)

            assert target_queue, 'Could not find queue to handle %r' % item

            if not item.fire_and_forget:
                self.pending[item] = barrier
            target_queue.put(item)

    def dequeue_barrier(self, item):
        # This is a WorkItem from a worker thread that has finished and
        # needs to be reinjected into a WorkflowItem generator.
        barrier = self.pending.pop(item, None)
        if not barrier:
            # Item was already finished in another barrier, or was
            # fire-and-forget and never part of a barrier; ignore it.
            return None

        if barrier.outstanding and not barrier.error:
            # More work to do and no error seen. Keep waiting.
            return None

        for work in barrier:
            # The barrier has been fulfilled one way or another. Clear out any
            # other pending parts of the barrier so they don't trigger again.
            self.pending.pop(work, None)

        return barrier

    def handle_item(self, item):
        if isinstance(item, WorkflowItem) and not item.done:
            workflow = item
            try:
                generator = item.run(*item.args, **item.kwargs)
            except TypeError, e:
                raise TypeError('Bad workflow function item=%r error=%s' % (
                                item, str(e)))
            item = None
        else:
            barrier = self.dequeue_barrier(item)
            if not barrier:
                logging.debug('Could not find barrier for finished item=%r',
                              item)
                return
            item = barrier.get_item()
            workflow = barrier.workflow
            generator = barrier.generator

        while True:
            try:
                try:
                    error = item is not None and item.error
                    if error:
                        logging.debug('Throwing workflow=%r error=%r',
                                      workflow, error)
                        next_item = generator.throw(*error)
                    elif isinstance(item, WorkflowItem) and item.done:
                        logging.debug(
                            'Sending workflow=%r finished item=%r',
                            workflow, item)
                        next_item = generator.send(item.result)
                    else:
                        logging.debug(
                            'Sending workflow=%r finished item=%r',
                            workflow, item)
                        next_item = generator.send(item)
                except StopIteration:
                    logging.debug('Exhausted workflow=%r', workflow)
                    workflow.done = True
                except Return, e:
                    logging.debug('Return from workflow=%r result=%r',
                                  workflow, e.result)
                    workflow.done = True
                    workflow.result = e.result
                except Exception, e:
                    logging.exception(
                        'Error in workflow=%r from item=%r, error=%r',
                        workflow, item, error)
                    workflow.done = True
                    workflow.error = sys.exc_info()
            finally:
                if workflow.done:
                    if workflow.root:
                        # Root workflow finished. This goes to the output
                        # queue so it can be received by the main thread.
                        return workflow
                    else:
                        # Sub-workflow finished. Reinject it into the
                        # workflow so a pending parent can catch it.
                        self.input_queue.put(workflow)
                        return

            barrier = Barrier(workflow, generator, next_item)
            self.enqueue_barrier(barrier)
            if barrier.outstanding:
                break

            # If a returned barrier has no oustanding parts, immediately
            # progress the workflow.
            item = barrier.get_item()


class PrintWorkflow(WorkflowItem):
    """Prints a message to stdout."""

    def run(self, message):
        yield []  # Make this into a generator
        print message


def get_coordinator():
    """Creates a coordinator and returns it."""
    workflow_queue = Queue.Queue()
    complete_queue = Queue.Queue()
    coordinator = WorkflowThread(workflow_queue, complete_queue)
    coordinator.register(WorkflowItem, workflow_queue)
    return coordinator



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements email sending for the API server and frontend."""

import logging

# Local libraries
from flask import render_template, request
from flask.ext.mail import Message
from flask.ext.login import current_user

# Local modules
from . import app
from . import mail
from dpxdt.server import models
from dpxdt.server import operations
from dpxdt.server import utils


def render_or_send(func, message):
    """Renders an email message for debugging or actually sends it."""
    if request.endpoint != func.func_name:
        mail.send(message)

    if (current_user.is_authenticated() and current_user.superuser):
        return render_template('debug_email.html', message=message)


@utils.ignore_exceptions
@app.route('/email/ready_for_review/<int:build_id>/'
           '<string:release_name>/<int:release_number>')
def send_ready_for_review(build_id, release_name, release_number):
    """Sends an email indicating that the release is ready for review."""
    build = models.Build.query.get(build_id)

    if not build.send_email:
        logging.debug(
            'Not sending ready for review email because build does not have '
            'email enabled. build_id=%r', build.id)
        return

    ops = operations.BuildOps(build_id)
    release, run_list, stats_dict, _ = ops.get_release(
        release_name, release_number)

    if not run_list:
        logging.debug(
            'Not sending ready for review email because there are '
            ' no runs. build_id=%r, release_name=%r, release_number=%d',
            build.id, release.name, release.number)
        return

    title = '%s: %s - Ready for review' % (build.name, release.name)

    email_body = render_template(
        'email_ready_for_review.html',
        build=build,
        release=release,
        run_list=run_list,
        stats_dict=stats_dict)

    recipients = []
    if build.email_alias:
        recipients.append(build.email_alias)
    else:
        for user in build.owners:
            recipients.append(user.email_address)

    if not recipients:
        logging.debug(
            'Not sending ready for review email because there are no '
            'recipients. build_id=%r, release_name=%r, release_number=%d',
            build.id, release.name, release.number)
        return

    message = Message(title, recipients=recipients)
    message.html = email_body

    logging.info('Sending ready for review email for build_id=%r, '
                 'release_name=%r, release_number=%d to %r',
                 build.id, release.name, release.number, recipients)

    return render_or_send(send_ready_for_review, message)



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Workers that consumer a release server's work queue."""

import logging

# Local Libraries
import gflags
FLAGS = gflags.FLAGS

# Local modules
from dpxdt.client import fetch_worker
from dpxdt.client import timer_worker
from dpxdt.client import workers


gflags.DEFINE_string(
    'queue_server_prefix', None,
    'URL prefix of where the work queue server is located, such as '
    '"https://www.example.com/api/work_queue". This should use HTTPS if '
    'possible, since API requests send credentials using HTTP basic auth.')

gflags.DEFINE_integer(
    'queue_idle_poll_seconds', 60,
    'How often to poll the work queue for new tasks when the worker is '
    'currently not processing any tasks.')

gflags.DEFINE_integer(
    'queue_busy_poll_seconds', 1,
    'How often to poll tasks running locally to see if they have completed '
    'and then go back to the server to look for more work.')


class Error(Exception):
    """Base-class for exceptions in this module."""


class GiveUpAfterAttemptsError(Error):
    """Exception indicates the task should give up after N attempts."""

    def __init__(self, max_attempts, *args, **kwargs):
        """Initializer.

        Args:
            max_attempts: Maximum number of attempts to make for this task,
                inclusive. So 2 means try two times and then retire the task.
            *args, **kwargs: Optional Exception arguments.
        """
        Exception.__init__(self, *args, **kwargs)
        self.max_attempts = max_attempts


class HeartbeatError(Error):
    """Reporting the status of a task in progress failed for some reason."""


class HeartbeatWorkflow(workers.WorkflowItem):
    """Reports the status of a RemoteQueueWorkflow to the API server.

    Args:
        queue_url: Base URL of the work queue.
        task_id: ID of the task to update the heartbeat status message for.
        message: Heartbeat status message to report.
        index: Index for the heartbeat message. Should be at least one
            higher than the last heartbeat message.
    """

    def run(self, queue_url, task_id, message, index):
        call = yield fetch_worker.FetchItem(
            queue_url + '/heartbeat',
            post={
                'task_id': task_id,
                'message': message,
                'index': index,
            },
            username=FLAGS.release_client_id,
            password=FLAGS.release_client_secret)

        if call.json and call.json.get('error'):
            raise HeartbeatError(call.json.get('error'))

        if not call.json or not call.json.get('success'):
            raise HeartbeatError('Bad response: %r' % call)


class DoTaskWorkflow(workers.WorkflowItem):
    """Runs a local workflow for a task and marks it done in the remote queue.

    Args:
        queue_url: Base URL of the work queue.
        local_queue_workflow: WorkflowItem sub-class to create using parameters
            from the remote work payload that will execute the task.
        task: JSON payload of the task.
        wait_seconds: Wait this many seconds before starting work.
            Defaults to zero.
    """

    fire_and_forget = True

    def run(self, queue_url, local_queue_workflow, task, wait_seconds=0):
        logging.info('Starting work item from queue_url=%r, '
                     'task=%r, workflow=%r, wait_seconds=%r',
                     queue_url, task, local_queue_workflow, wait_seconds)

        if wait_seconds > 0:
            yield timer_worker.TimerItem(wait_seconds)

        # Define a heartbeat closure that will return a workflow for
        # reporting status. This will auto-increment the index on each
        # call, so only the latest update will be saved.
        index = [0]
        task_id = task['task_id']
        def heartbeat(message):
            next_index = index[0]
            index[0] = next_index + 1
            return HeartbeatWorkflow(
                queue_url, task_id, message, next_index)

        payload = task['payload']
        payload.update(heartbeat=heartbeat)

        error = False

        try:
            yield local_queue_workflow(**payload)
        except Exception, e:
            logging.exception('Exception while processing work from '
                              'queue_url=%r, task=%r', queue_url, task)
            yield heartbeat('%s: %s' % (e.__class__.__name__, str(e)))

            if (isinstance(e, GiveUpAfterAttemptsError) and
                    task['lease_attempts'] >= e.max_attempts):
                logging.warning(
                    'Hit max attempts on task=%r, marking task as error',
                    task)
                error = True
            else:
                # The task has legimiately failed. Do not mark the task as
                # finished. Let it retry in the queue again.
                return

        finish_params = {'task_id': task_id}
        if error:
            finish_params['error'] = '1'

        try:
            finish_item = yield fetch_worker.FetchItem(
                queue_url + '/finish',
                post=finish_params,
                username=FLAGS.release_client_id,
                password=FLAGS.release_client_secret)
        except Exception, e:
            logging.error('Could not finish work with '
                          'queue_url=%r, task=%r. %s: %s',
                          queue_url, task, e.__class__.__name__, e)
        else:
            if finish_item.json and finish_item.json.get('error'):
                logging.error('Could not finish work with '
                              'queue_url=%r, task=%r. %s',
                              queue_url, finish_item.json['error'], task)
            else:
                logging.info('Finished work item with queue_url=%r, '
                             'task_id=%r', queue_url, task_id)


class RemoteQueueWorkflow(workers.WorkflowItem):
    """Fetches tasks from a remote queue periodically, runs them locally.

    Args:
        queue_name: Name of the queue to fetch from.
        local_queue_workflow: WorkflowItem sub-class to create using parameters
            from the remote work payload that will execute the task.
        max_tasks: Maximum number of tasks to have in flight at any time.
            Defaults to 1.
        wait_seconds: How many seconds should be between tasks starting to
            process locally. Defaults to 0. Can be used to spread out
            the load a new set of tasks has on the server.
    """

    def run(self, queue_name, local_queue_workflow,
            max_tasks=1, wait_seconds=0):
        queue_url = '%s/%s' % (FLAGS.queue_server_prefix, queue_name)
        outstanding = []

        while True:
            next_count = max_tasks - len(outstanding)
            next_tasks = []

            if next_count > 0:
                logging.info(
                    'Fetching %d tasks from queue_url=%r for workflow=%r',
                    next_count, queue_url, local_queue_workflow)
                try:
                    next_item = yield fetch_worker.FetchItem(
                        queue_url + '/lease',
                        post={'count': next_count},
                        username=FLAGS.release_client_id,
                        password=FLAGS.release_client_secret)
                except Exception, e:
                    logging.error(
                        'Could not fetch work from queue_url=%r. %s: %s',
                        queue_url, e.__class__.__name__, e)
                else:
                    if next_item.json:
                        if next_item.json.get('error'):
                            logging.error(
                                'Could not fetch work from queue_url=%r. %s',
                                queue_url, next_item.json['error'])
                        elif next_item.json['tasks']:
                            next_tasks = next_item.json['tasks']

            for index, task in enumerate(next_tasks):
                item = yield DoTaskWorkflow(
                    queue_url, local_queue_workflow, task,
                    wait_seconds=index * wait_seconds)
                outstanding.append(item)

            # Poll for new tasks frequently when we're currently handling
            # task load. Poll infrequently when there hasn't been anything
            # to do recently.
            poll_time = FLAGS.queue_idle_poll_seconds
            if outstanding:
                poll_time = FLAGS.queue_busy_poll_seconds

            yield timer_worker.TimerItem(poll_time)

            outstanding[:] = [x for x in outstanding if not x.done]
            logging.debug('%d items for %r still outstanding: %r',
                          len(outstanding), local_queue_workflow, outstanding)



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Frontend for the API server."""

import base64
import datetime
import hashlib
import logging

# Local libraries
import flask
from flask import Flask, abort, g, redirect, render_template, request, url_for
from flask.ext.login import (
    current_user, fresh_login_required, login_fresh, login_required)
from flask.ext.wtf import Form

# Local modules
from . import app
from . import db
from . import login
from dpxdt.server import auth
from dpxdt.server import forms
from dpxdt.server import models
from dpxdt.server import operations
from dpxdt.server import signals
from dpxdt.server import utils


@app.context_processor
def frontend_context():
    """Adds extra default context for rendered templates."""
    return dict(cache_buster=utils.get_deployment_timestamp())


@app.route('/')
def homepage():
    """Renders the homepage."""
    if current_user.is_authenticated():
        if not login_fresh():
            logging.debug('User needs a fresh token')
            abort(login.needs_refresh())

        auth.claim_invitations(current_user)

    build_list = operations.UserOps(current_user.get_id()).get_builds()

    return render_template(
        'home.html',
        build_list=build_list)


@app.route('/new', methods=['GET', 'POST'])
@fresh_login_required
def new_build():
    """Page for crediting or editing a build."""
    form = forms.BuildForm()
    if form.validate_on_submit():
        build = models.Build()
        form.populate_obj(build)
        build.owners.append(current_user)

        db.session.add(build)
        db.session.flush()

        auth.save_admin_log(build, created_build=True, message=build.name)

        db.session.commit()

        operations.UserOps(current_user.get_id()).evict()

        logging.info('Created build via UI: build_id=%r, name=%r',
                     build.id, build.name)
        return redirect(url_for('view_build', id=build.id))

    return render_template(
        'new_build.html',
        build_form=form)


@app.route('/build')
@auth.build_access_required
def view_build():
    """Page for viewing all releases in a build."""
    build = g.build
    page_size = 20
    offset = request.args.get('offset', 0, type=int)

    ops = operations.BuildOps(build.id)
    has_next_page, candidate_list, stats_counts = ops.get_candidates(
        page_size, offset)

    # Collate by release name, order releases by latest creation. Init stats.
    release_dict = {}
    created_dict = {}
    run_stats_dict = {}
    for candidate in candidate_list:
        release_list = release_dict.setdefault(candidate.name, [])
        release_list.append(candidate)
        max_created = created_dict.get(candidate.name, candidate.created)
        created_dict[candidate.name] = max(candidate.created, max_created)
        run_stats_dict[candidate.id] = dict(
            runs_total=0,
            runs_complete=0,
            runs_successful=0,
            runs_failed=0,
            runs_baseline=0,
            runs_pending=0)

    # Sort each release by candidate number descending
    for release_list in release_dict.itervalues():
        release_list.sort(key=lambda x: x.number, reverse=True)

    # Sort all releases by created time descending
    release_age_list = [
        (value, key) for key, value in created_dict.iteritems()]
    release_age_list.sort(reverse=True)
    release_name_list = [key for _, key in release_age_list]

    # Count totals for each run state within that release.
    for candidate_id, status, count in stats_counts:
        stats_dict = run_stats_dict[candidate_id]
        for key in ops.get_stats_keys(status):
            stats_dict[key] += count

    return render_template(
        'view_build.html',
        build=build,
        release_name_list=release_name_list,
        release_dict=release_dict,
        run_stats_dict=run_stats_dict,
        has_next_page=has_next_page,
        current_offset=offset,
        next_offset=offset + page_size,
        last_offset=max(0, offset -  page_size))


@app.route('/release', methods=['GET', 'POST'])
@auth.build_access_required
def view_release():
    """Page for viewing all tests runs in a release."""
    build = g.build
    if request.method == 'POST':
        form = forms.ReleaseForm(request.form)
    else:
        form = forms.ReleaseForm(request.args)

    form.validate()

    ops = operations.BuildOps(build.id)
    release, run_list, stats_dict, approval_log = ops.get_release(
        form.name.data, form.number.data)

    if not release:
        abort(404)

    if request.method == 'POST':
        decision_states = (
            models.Release.REVIEWING,
            models.Release.RECEIVING,
            models.Release.PROCESSING)

        if form.good.data and release.status in decision_states:
            release.status = models.Release.GOOD
            auth.save_admin_log(build, release_good=True, release=release)
        elif form.bad.data and release.status in decision_states:
            release.status = models.Release.BAD
            auth.save_admin_log(build, release_bad=True, release=release)
        elif form.reviewing.data and release.status in (
                models.Release.GOOD, models.Release.BAD):
            release.status = models.Release.REVIEWING
            auth.save_admin_log(build, release_reviewing=True, release=release)
        else:
            logging.warning(
                'Bad state transition for name=%r, number=%r, form=%r',
                release.name, release.number, form.data)
            abort(400)

        db.session.add(release)
        db.session.commit()

        ops.evict()

        return redirect(url_for(
            'view_release',
            id=build.id,
            name=release.name,
            number=release.number))

    # Update form values for rendering
    form.good.data = True
    form.bad.data = True
    form.reviewing.data = True

    return render_template(
        'view_release.html',
        build=build,
        release=release,
        run_list=run_list,
        release_form=form,
        approval_log=approval_log,
        stats_dict=stats_dict)


def _get_artifact_context(run, file_type):
    """Gets the artifact details for the given run and file_type."""
    sha1sum = None
    image_file = False
    log_file = False
    config_file = False

    if request.path == '/image':
        image_file = True
        if file_type == 'before':
            sha1sum = run.ref_image
        elif file_type == 'diff':
            sha1sum = run.diff_image
        elif file_type == 'after':
            sha1sum = run.image
        else:
            abort(400)
    elif request.path == '/log':
        log_file = True
        if file_type == 'before':
            sha1sum = run.ref_log
        elif file_type == 'diff':
            sha1sum = run.diff_log
        elif file_type == 'after':
            sha1sum = run.log
        else:
            abort(400)
    elif request.path == '/config':
        config_file = True
        if file_type == 'before':
            sha1sum = run.ref_config
        elif file_type == 'after':
            sha1sum = run.config
        else:
            abort(400)

    return image_file, log_file, config_file, sha1sum


@app.route('/run', methods=['GET', 'POST'])
@app.route('/image', endpoint='view_image', methods=['GET', 'POST'])
@app.route('/log', endpoint='view_log', methods=['GET', 'POST'])
@app.route('/config', endpoint='view_config', methods=['GET', 'POST'])
@auth.build_access_required
def view_run():
    """Page for viewing before/after for a specific test run."""
    build = g.build
    if request.method == 'POST':
        form = forms.RunForm(request.form)
    else:
        form = forms.RunForm(request.args)

    form.validate()

    ops = operations.BuildOps(build.id)
    run, next_run, previous_run, approval_log = ops.get_run(
        form.name.data, form.number.data, form.test.data)

    if not run:
        abort(404)

    file_type = form.type.data
    image_file, log_file, config_file, sha1sum = (
        _get_artifact_context(run, file_type))

    if request.method == 'POST':
        if form.approve.data and run.status == models.Run.DIFF_FOUND:
            run.status = models.Run.DIFF_APPROVED
            auth.save_admin_log(build, run_approved=True, run=run)
        elif form.disapprove.data and run.status == models.Run.DIFF_APPROVED:
            run.status = models.Run.DIFF_FOUND
            auth.save_admin_log(build, run_rejected=True, run=run)
        else:
            abort(400)

        db.session.add(run)
        db.session.commit()

        ops.evict()

        return redirect(url_for(
            request.endpoint,
            id=build.id,
            name=run.release.name,
            number=run.release.number,
            test=run.name,
            type=file_type))

    # Update form values for rendering
    form.approve.data = True
    form.disapprove.data = True

    context = dict(
        build=build,
        release=run.release,
        run=run,
        run_form=form,
        previous_run=previous_run,
        next_run=next_run,
        file_type=file_type,
        image_file=image_file,
        log_file=log_file,
        config_file=config_file,
        sha1sum=sha1sum,
        approval_log=approval_log)

    if file_type:
        template_name = 'view_artifact.html'
    else:
        template_name = 'view_run.html'

    response = flask.Response(render_template(template_name, **context))

    return response


@app.route('/settings', methods=['GET', 'POST'])
@auth.build_access_required('build_id')
def build_settings():
    build = g.build

    settings_form = forms.SettingsForm()

    if settings_form.validate_on_submit():
        settings_form.populate_obj(build)

        message = ('name=%s, send_email=%s, email_alias=%s' % (
            build.name, build.send_email, build.email_alias))
        auth.save_admin_log(build, changed_settings=True, message=message)

        db.session.add(build)
        db.session.commit()

        signals.build_updated.send(app, build=build, user=current_user)

        return redirect(url_for(
            request.endpoint,
            build_id=build.id))

    # Update form values for rendering
    settings_form.name.data = build.name
    settings_form.build_id.data = build.id
    settings_form.email_alias.data = build.email_alias
    settings_form.send_email.data = build.send_email

    return render_template(
        'view_settings.html',
        build=build,
        settings_form=settings_form)
